import threading
import types
import typing as t

Callback = t.Callable[[], None]


class Event(threading.Event):
    """
    A threading.Event that also calls back to zero or more functions when its state
    is set or reset, and has a __bool__ method.

    Note that the callback might happen on some completely different thread,
    so these functions cannot block"""

    on_set: t.List[Callback]

    def __init__(self, *on_set: Callback) -> None:
        self.on_set = list(on_set)
        super().__init__()

    def set(self) -> None:
        super().set()
        [c() for c in self.on_set]

    def clear(self) -> None:
        super().clear()
        [c() for c in self.on_set]

    def __bool__(self) -> bool:
        return self.is_set()


class _Event:
    def __set_name__(self, owner: t.Any, name: str) -> None:
        self.name = '_' + name

    def __get__(self, obj: t.Any, objtype: t.Any = None) -> Event:
        event: t.Optional[Event] = getattr(obj, self.name, None)
        if event is not None:
            return event

        setattr(obj, self.name, event := Event())
        return event

    def __set__(self, obj: t.Any, value: t.Union[bool, Event]) -> None:
        if isinstance(value, Event):
            setattr(obj, self.name, value)
        elif value:
            self.__get__(obj).set()
        else:
            self.__get__(obj).clear()


class Runnable:
    """A base class for things that start, run, finish, stop and join

    Stopping is requesting immediate termination: finishing is saying that
    there is no more work to be done, finish what you are doing.

    A Runnable has two `Event`s, `running` and `stopped`, and you can either
    `wait` on either of these conditions to be true, or add a callback function
    (which must be non-blocking) to either of them.

    `running` is not set until the setup for a `Runnable` has finished;
    `stopped` is not set until all the computations in a thread have ceased.

    An Runnable can be used as a context manager:

        with runnable:
            # The runnable is running by this point
            do_stuff()
        # By the time you get to here, the runnable has completely stopped

    The above means roughly the same as

        runnable.start()
        try:
            do_stuff()
            runnable.finish()
            runnable.join()
        finally:
            runnable.stop()

    """

    running = _Event()
    stopped = _Event()

    #: Do we join the thread in __exit__?
    join_on_exit: bool = True

    def start(self) -> None:
        """Start this object.

        Note that self.running might not be immediately true after this method completes
        """
        self.running = True

    def stop(self) -> None:
        """Stop as soon as possible. might not do anything, should never raise.

        Note that self.stopped might not be immediately true after this method completes
        """
        self.running = False
        self.stopped = True

    def finish(self) -> None:
        """Request an orderly shutdown where all existing work is completed.

        Note that self.stopped might not be immediately true after this method completes
        """
        self.stop()

    def join(self, timeout: t.Optional[float] = None) -> None:
        """Join this thread or process.  Might block indefinitely, might do nothing"""

    def __enter__(self) -> 'Runnable':
        self.start()
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[types.TracebackType],
    ) -> None:
        if exc_type in (None, KeyboardInterrupt):
            self.finish()
            if self.join_on_exit:
                self.join()
        else:
            self.stop()



from .collection import HasRunnables, Runnables, ThreadQueue
from .runnable import Event, Runnable
from .thread import HasThread, IsThread
from .wrapper import Wrapper

__all__ = (
    'Event',
    'HasRunnables',
    'HasThread',
    'IsThread',
    'Runnable',
    'Runnables',
    'ThreadQueue',
    'Wrapper',
)



import typing as t

from .runnable import Runnable


class Wrapper(Runnable):
    """Duck-type wrapper for runnable things"""

    def __init__(self, wrapped: t.Any) -> None:
        self.wrapped = wrapped
        super().__init__()

    def start(self) -> None:
        super().start()
        if (start := getattr(self.wrapped, 'start', None)) is not None:
            start()

    def stop(self) -> None:
        super().stop()
        if (stop := getattr(self.wrapped, 'stop', None)) is not None:
            stop()

    def join(self, timeout: t.Optional[float] = None) -> None:
        super().join(timeout)
        if (join := getattr(self.wrapped, 'join', None)) is not None:
            join(timeout)



import dataclasses as dc
import functools
import logging
import traceback
import typing as t
from threading import Thread
from time import sleep

from .runnable import Callback, Runnable

__all__ = 'ExceptionHandler', 'HasThread', 'IsLog', 'IsThread'

ExceptionHandler = t.Callable[[Exception], None]


class AnyFunction(t.Protocol):
    def __call__(self, *args: t.Any, **kwargs: t.Any) -> t.Any:
        pass


class IsLog(t.Protocol):
    """A class that looks like a subset of Python's logging"""

    debug: AnyFunction
    error: AnyFunction


class HasLog(t.Protocol):
    log: IsLog


def _add_log(
    method: t.Optional[AnyFunction] = None, before: bool = True, after: bool = False
) -> AnyFunction:
    if method is None:
        return functools.partial(_add_log, before=before, after=after)

    name = getattr(method, '__name__', str(method))

    @functools.wraps(method)
    def wrapped(self: HasLog, *a: t.Any, **ka: t.Any) -> t.Any:
        msg = f'{self}: {name}'

        try:
            if before:
                self.log.debug(f'{msg}: before')
            return method(self, *a, **ka)
        finally:
            if after:
                self.log.debug(f'{msg}: after')

    return wrapped


class ThreadBase(Runnable):
    """A base class for classes with a thread.

    It adds the following features to threading.Thread:

    * Has Events `running` and `stopped` with `on_set` callbacks
    * Handles exceptions and prints or redirects them
    * Runs once, or multiple times, depending on `self.looping`
    """

    #: `callback` is called one or more times in a new thread, depending
    #: on the value of `looping` and whether the thread gets stopped
    callback: Callback

    #: If set, `exception` gets called on an Exception.
    exception: t.Optional[ExceptionHandler] = None

    #: Passed to threading.Thread
    daemon: bool = False

    #: Do we join the thread in __exit__?
    join_on_exit: bool = True

    #: Used for error and debug logging
    log: IsLog = logging

    #: If True, `callback` is called repeatedly until `running` is unset
    #: If False, `callback` is called just once
    looping: bool = False

    #: The print name of the thread, used for debugging
    name: str = ''

    #: How long to delay after running?
    post_delay: float = 0

    #: How long to delay before running?
    pre_delay: float = 0

    def __str__(self) -> str:
        return f'({self.__class__.__name__}){self.name}'

    @_add_log
    def pre_run(self) -> None:
        pass

    @_add_log(after=True)
    def run(self) -> None:
        def tb() -> None:
            exc = traceback.format_exc()
            self.log.error(f'{self}: Exception\n{exc}')
            self.stop()

        try:
            self.pre_run()
        except Exception:
            tb()
            return

        self.running.set()

        while self.running:
            if self.pre_delay > 0:
                sleep(self.pre_delay)
            try:
                self.callback()
            except Exception as e:
                if self.exception is not None:
                    try:
                        self.exception(e)
                    except Exception:
                        tb()
                    else:
                        self.stop()
                else:
                    tb()
            else:
                if not self.looping:
                    self.stop()
                if self.post_delay > 0:
                    sleep(self.post_delay)

        self.stopped.set()

    @_add_log
    def stop(self) -> None:
        self.running.clear()

    @_add_log
    def finish(self) -> None:
        super().finish()


class IsThread(ThreadBase, Thread):
    """This ThreadBase inherits from threading.Thread.

    To use IsThread, derive from it and override either or both of
    self.callback() and self.pre_run()
    """

    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        ThreadBase.__init__(self, *args, **kwargs)
        Thread.__init__(self, daemon=self.daemon)

    def callback(self) -> None:
        """Called one or more times in a new thread"""

    @_add_log(after=True)
    def join(self, timeout: t.Optional[float] = None) -> None:
        Thread.join(self, timeout)

    @_add_log
    def start(self) -> None:
        Thread.start(self)


@dc.dataclass
class HasThread(ThreadBase):
    """This ThreadBase contains a thread, and is constructed with a callback"""

    #: `callback` is called one or more times in a new thread, depending
    #: on the value of `looping` and whether the thread gets stopped
    callback: Callback

    #: If set, `exception` gets called on an Exception.
    exception: t.Optional[ExceptionHandler] = None

    #: Passed to threading.Thread
    daemon: bool = False

    #: Do we join the thread in __exit__?
    join_on_exit: bool = True

    #: Used for error and debug logging
    log: IsLog = logging

    #: If True, `callback` is called repeated until `running` is unset
    #: If False, `callback` is called just once
    looping: bool = False

    #: The print name of the string, used for debugging
    name: str = ''

    #: How long to delay after running?
    post_delay: float = 0

    #: How long to delay before running?
    pre_delay: float = 0

    def __post_init__(self) -> None:
        ThreadBase.__init__(self)

    @_add_log(after=True)
    def join(self, timeout: t.Optional[float] = None) -> None:
        self.thread.join(timeout)

    @_add_log
    def start(self) -> None:
        self.thread.start()

    def new_thread(self) -> Thread:
        return Thread(target=self.run, daemon=self.daemon)

    @functools.cached_property
    def thread(self) -> Thread:
        return self.new_thread()



import dataclasses as dc
import logging
import typing as t
from functools import cached_property
from queue import Empty, Queue

from .runnable import Runnable
from .thread import ExceptionHandler, HasThread, IsLog

_SENTINEL_MESSAGE = object()
T = t.TypeVar('T')


class HasRunnables(Runnable):
    """Collect zero or more Runnable into one"""

    runnables: t.Sequence[Runnable]

    def start(self) -> None:
        for r in self.runnables:
            r.running.on_set.append(self._on_start)
            r.stopped.on_set.append(self._on_stop)
            r.start()

    def stop(self) -> None:
        self.running.clear()
        for r in self.runnables:
            r.stop()

    def finish(self) -> None:
        for r in self.runnables:
            r.finish()

    def join(self, timeout: t.Optional[float] = None) -> None:
        for r in self.runnables:
            r.join(timeout)

    def _on_start(self) -> None:
        if not self.running and all(r.running for r in self.runnables):
            super().start()

    def _on_stop(self) -> None:
        if not self.stopped and all(r.stopped for r in self.runnables):
            super().stop()


class Runnables(HasRunnables):
    def __init__(self, *runnables: Runnable):
        self.runnables = runnables
        super().__init__()


@dc.dataclass
class ThreadQueue(HasRunnables, t.Generic[T]):
    """A simple multi-producer, multi-consumer queue with one thread per consumer.

    There is a special `finish_message` value, which when received shuts down
    that consumer.  ThreadQueue.finish() puts one `self.finish_message` onto the
    queue for each consumer.
    """

    #: `callback` is called on one of the worker threads for each entry
    #: that gets added to the queue
    callback: t.Callable[[T], None]

    #: Passed to threading.Thread
    daemon: bool = False

    #: If set, `exception` gets called on an Exception.
    exception: t.Optional[ExceptionHandler] = None

    #: Do we join the thread in __exit__?
    join_on_exit: bool = True

    #: Used for error and debug logging
    log: IsLog = logging

    #: Passed to queue.Queue
    maxsize: int = 0

    #: The print name of the thread, used for debugging
    name: str = 'thread_queue'

    #: Number of threads to service the queue
    thread_count: int = 1

    #: Timeout in polling the queue
    timeout: t.Optional[float] = 0.1

    def __post_init__(self) -> None:
        def thread(i: int) -> HasThread:
            return HasThread(
                callback=self._callback,
                exception=self.exception,
                join_on_exit=self.join_on_exit,
                name=f'{self.name}-{i}',
            )

        HasRunnables.__init__(self)
        self.runnables = tuple(thread(i) for i in range(self.thread_count))

    def put(self, item: T) -> None:
        self.queue.put(item)

    @cached_property
    def queue(self) -> 'Queue[T]':
        # See https://stackoverflow.com/a/57728797/43839
        return Queue(self.maxsize)

    def finish(self) -> None:
        """Put an empty message into the queue for each listener"""
        for _ in self.runnables:
            self.put(t.cast(T, _SENTINEL_MESSAGE))

        super().finish()

    def _callback(self) -> None:
        self.running.wait()

        while self.running:
            try:
                item = self.queue.get(timeout=self.timeout)
            except Empty:
                continue
            if item == _SENTINEL_MESSAGE:
                return
            try:
                self.callback(item)
            except Exception:
                self.stop()
                raise
